
/* JavaScript content from wlclient/js/encryptedcache/encryptedcache.js in Common Resources */
/*
 * Licensed Materials - Property of IBM
 * 5725-G92 (C) Copyright IBM Corp. 2006, 2013. All Rights Reserved.
 * US Government Users Restricted Rights - Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 */

/**
 * Encrypted Off-line Cache
 */
/*
 * Use local variables for non-obfuscatable variables to prevent static code inspection
 */

var myMathRandom = Math.random;
var myMathFloor = Math.floor;
var myMathCeil = Math.ceil;
var myStringFromCharCode = String.fromCharCode;
var myDecodeURIComponent = decodeURIComponent;
var myEncodeURIComponent = encodeURIComponent;
var myUnescape = unescape;
var myEscape = escape;

var api = {
    open: "open",
    close: "close",
    changeCredentials: "changeCredentials",
    destroy: "destroy",
    read: "read",
    write: "write",
    remove: "remove",
    secureRandom: "secureRandom",
    random: "random",
    keygen: "keygen"
};

function a2h(numArr) {
    var string = '',
        i;
    for (i = 0; i < numArr.length; i++) {
        string += (numArr[i] < 16 ? '0' : '') + numArr[i].toString(16);
    }
    return string;
}

function h2a(s) {
    var ret = [];
    s.replace(/(..)/g, function(s) {
        ret.push(parseInt(s, 16));
    });
    return ret;
}

function s2a(string) {
    var array = [],
        i;
    for (i = 0; i < string.length; i++) {
        array[i] = string.charCodeAt(i);
    }
    return array;
}

function enc_utf8(s) {
    try {
        return myUnescape(myEncodeURIComponent(s));
    } catch (e) {
        throw e;
    }
}

function dec_utf8(s) {
    try {
        return myDecodeURIComponent(myEscape(s));
    } catch (e) {
        throw e;
    }
}

function MD5(numArr) {

    function rotateLeft(lValue, iShiftBits) {
        return (lValue << iShiftBits) | (lValue >>> (32 - iShiftBits));
    }

    function addUnsigned(lX, lY) {
        var lX4, lY4, lX8, lY8, lResult;
        lX8 = (lX & 0x80000000);
        lY8 = (lY & 0x80000000);
        lX4 = (lX & 0x40000000);
        lY4 = (lY & 0x40000000);
        lResult = (lX & 0x3FFFFFFF) + (lY & 0x3FFFFFFF);
        if (lX4 & lY4) {
            return (lResult ^ 0x80000000 ^ lX8 ^ lY8);
        }
        if (lX4 | lY4) {
            if (lResult & 0x40000000) {
                return (lResult ^ 0xC0000000 ^ lX8 ^ lY8);
            } else {
                return (lResult ^ 0x40000000 ^ lX8 ^ lY8);
            }
        } else {
            return (lResult ^ lX8 ^ lY8);
        }
    }

    function f(x, y, z) {
        return (x & y) | ((~x) & z);
    }

    function g(x, y, z) {
        return (x & z) | (y & (~z));
    }

    function h(x, y, z) {
        return (x ^ y ^ z);
    }

    function funcI(x, y, z) {
        return (y ^ (x | (~z)));
    }

    function ff(a, b, c, d, x, s, ac) {
        a = addUnsigned(a, addUnsigned(addUnsigned(f(b, c, d), x), ac));
        return addUnsigned(rotateLeft(a, s), b);
    }

    function gg(a, b, c, d, x, s, ac) {
        a = addUnsigned(a, addUnsigned(addUnsigned(g(b, c, d), x), ac));
        return addUnsigned(rotateLeft(a, s), b);
    }

    function hh(a, b, c, d, x, s, ac) {
        a = addUnsigned(a, addUnsigned(addUnsigned(h(b, c, d), x), ac));
        return addUnsigned(rotateLeft(a, s), b);
    }

    function ii(a, b, c, d, x, s, ac) {
        a = addUnsigned(a, addUnsigned(addUnsigned(funcI(b, c, d), x), ac));
        return addUnsigned(rotateLeft(a, s), b);
    }

    function convertToWordArray(numArr) {
        var lWordCount, lMessageLength = numArr.length,
            lNumberOfWords_temp1 = lMessageLength + 8,
            lNumberOfWords_temp2 = (lNumberOfWords_temp1 - (lNumberOfWords_temp1 % 64)) / 64,
            lNumberOfWords = (lNumberOfWords_temp2 + 1) * 16,
            lWordArray = [],
            lBytePosition = 0,
            lByteCount = 0;
        while (lByteCount < lMessageLength) {
            lWordCount = (lByteCount - (lByteCount % 4)) / 4;
            lBytePosition = (lByteCount % 4) * 8;
            lWordArray[lWordCount] = (lWordArray[lWordCount] | (numArr[lByteCount] << lBytePosition));
            lByteCount++;
        }
        lWordCount = (lByteCount - (lByteCount % 4)) / 4;
        lBytePosition = (lByteCount % 4) * 8;
        lWordArray[lWordCount] = lWordArray[lWordCount] | (0x80 << lBytePosition);
        lWordArray[lNumberOfWords - 2] = lMessageLength << 3;
        lWordArray[lNumberOfWords - 1] = lMessageLength >>> 29;
        return lWordArray;
    }

    function wordToHex(lValue) {
        var lByte, lCount, wordToHexArr = [];
        for (lCount = 0; lCount <= 3; lCount++) {
            lByte = (lValue >>> (lCount * 8)) & 255;
            wordToHexArr = wordToHexArr.concat(lByte);
        }
        return wordToHexArr;
    }

    /*
     * function utf8Encode(string) { string = string.replace(/\r\n/g, "\n"); var utftext = "", n, c;
     *
     * for (n = 0; n < string.length; n++) {
     *
     * c = string.charCodeAt(n);
     *
     * if (c < 128) { utftext += String.fromCharCode(c); } else if ((c > 127) && (c < 2048)) { utftext +=
     * String.fromCharCode((c >> 6) | 192); utftext += String.fromCharCode((c & 63) | 128); } else {
     * utftext += String.fromCharCode((c >> 12) | 224); utftext += String.fromCharCode(((c >> 6) & 63) |
     * 128); utftext += String.fromCharCode((c & 63) | 128); } }
     *
     * return utftext; }
     */

    var x = [],
        k, AA, BB, CC, DD, a, b, c, d, S11 = 7,
        S12 = 12,
        S13 = 17,
        S14 = 22,
        S21 = 5,
        S22 = 9,
        S23 = 14,
        S24 = 20,
        S31 = 4,
        S32 = 11,
        S33 = 16,
        S34 = 23,
        S41 = 6,
        S42 = 10,
        S43 = 15,
        S44 = 21;

    x = convertToWordArray(numArr);

    a = 0x67452301;
    b = 0xEFCDAB89;
    c = 0x98BADCFE;
    d = 0x10325476;

    for (k = 0; k < x.length; k += 16) {
        AA = a;
        BB = b;
        CC = c;
        DD = d;
        a = ff(a, b, c, d, x[k + 0], S11, 0xD76AA478);
        d = ff(d, a, b, c, x[k + 1], S12, 0xE8C7B756);
        c = ff(c, d, a, b, x[k + 2], S13, 0x242070DB);
        b = ff(b, c, d, a, x[k + 3], S14, 0xC1BDCEEE);
        a = ff(a, b, c, d, x[k + 4], S11, 0xF57C0FAF);
        d = ff(d, a, b, c, x[k + 5], S12, 0x4787C62A);
        c = ff(c, d, a, b, x[k + 6], S13, 0xA8304613);
        b = ff(b, c, d, a, x[k + 7], S14, 0xFD469501);
        a = ff(a, b, c, d, x[k + 8], S11, 0x698098D8);
        d = ff(d, a, b, c, x[k + 9], S12, 0x8B44F7AF);
        c = ff(c, d, a, b, x[k + 10], S13, 0xFFFF5BB1);
        b = ff(b, c, d, a, x[k + 11], S14, 0x895CD7BE);
        a = ff(a, b, c, d, x[k + 12], S11, 0x6B901122);
        d = ff(d, a, b, c, x[k + 13], S12, 0xFD987193);
        c = ff(c, d, a, b, x[k + 14], S13, 0xA679438E);
        b = ff(b, c, d, a, x[k + 15], S14, 0x49B40821);
        a = gg(a, b, c, d, x[k + 1], S21, 0xF61E2562);
        d = gg(d, a, b, c, x[k + 6], S22, 0xC040B340);
        c = gg(c, d, a, b, x[k + 11], S23, 0x265E5A51);
        b = gg(b, c, d, a, x[k + 0], S24, 0xE9B6C7AA);
        a = gg(a, b, c, d, x[k + 5], S21, 0xD62F105D);
        d = gg(d, a, b, c, x[k + 10], S22, 0x2441453);
        c = gg(c, d, a, b, x[k + 15], S23, 0xD8A1E681);
        b = gg(b, c, d, a, x[k + 4], S24, 0xE7D3FBC8);
        a = gg(a, b, c, d, x[k + 9], S21, 0x21E1CDE6);
        d = gg(d, a, b, c, x[k + 14], S22, 0xC33707D6);
        c = gg(c, d, a, b, x[k + 3], S23, 0xF4D50D87);
        b = gg(b, c, d, a, x[k + 8], S24, 0x455A14ED);
        a = gg(a, b, c, d, x[k + 13], S21, 0xA9E3E905);
        d = gg(d, a, b, c, x[k + 2], S22, 0xFCEFA3F8);
        c = gg(c, d, a, b, x[k + 7], S23, 0x676F02D9);
        b = gg(b, c, d, a, x[k + 12], S24, 0x8D2A4C8A);
        a = hh(a, b, c, d, x[k + 5], S31, 0xFFFA3942);
        d = hh(d, a, b, c, x[k + 8], S32, 0x8771F681);
        c = hh(c, d, a, b, x[k + 11], S33, 0x6D9D6122);
        b = hh(b, c, d, a, x[k + 14], S34, 0xFDE5380C);
        a = hh(a, b, c, d, x[k + 1], S31, 0xA4BEEA44);
        d = hh(d, a, b, c, x[k + 4], S32, 0x4BDECFA9);
        c = hh(c, d, a, b, x[k + 7], S33, 0xF6BB4B60);
        b = hh(b, c, d, a, x[k + 10], S34, 0xBEBFBC70);
        a = hh(a, b, c, d, x[k + 13], S31, 0x289B7EC6);
        d = hh(d, a, b, c, x[k + 0], S32, 0xEAA127FA);
        c = hh(c, d, a, b, x[k + 3], S33, 0xD4EF3085);
        b = hh(b, c, d, a, x[k + 6], S34, 0x4881D05);
        a = hh(a, b, c, d, x[k + 9], S31, 0xD9D4D039);
        d = hh(d, a, b, c, x[k + 12], S32, 0xE6DB99E5);
        c = hh(c, d, a, b, x[k + 15], S33, 0x1FA27CF8);
        b = hh(b, c, d, a, x[k + 2], S34, 0xC4AC5665);
        a = ii(a, b, c, d, x[k + 0], S41, 0xF4292244);
        d = ii(d, a, b, c, x[k + 7], S42, 0x432AFF97);
        c = ii(c, d, a, b, x[k + 14], S43, 0xAB9423A7);
        b = ii(b, c, d, a, x[k + 5], S44, 0xFC93A039);
        a = ii(a, b, c, d, x[k + 12], S41, 0x655B59C3);
        d = ii(d, a, b, c, x[k + 3], S42, 0x8F0CCC92);
        c = ii(c, d, a, b, x[k + 10], S43, 0xFFEFF47D);
        b = ii(b, c, d, a, x[k + 1], S44, 0x85845DD1);
        a = ii(a, b, c, d, x[k + 8], S41, 0x6FA87E4F);
        d = ii(d, a, b, c, x[k + 15], S42, 0xFE2CE6E0);
        c = ii(c, d, a, b, x[k + 6], S43, 0xA3014314);
        b = ii(b, c, d, a, x[k + 13], S44, 0x4E0811A1);
        a = ii(a, b, c, d, x[k + 4], S41, 0xF7537E82);
        d = ii(d, a, b, c, x[k + 11], S42, 0xBD3AF235);
        c = ii(c, d, a, b, x[k + 2], S43, 0x2AD7D2BB);
        b = ii(b, c, d, a, x[k + 9], S44, 0xEB86D391);
        a = addUnsigned(a, AA);
        b = addUnsigned(b, BB);
        c = addUnsigned(c, CC);
        d = addUnsigned(d, DD);
    }

    return wordToHex(a).concat(wordToHex(b), wordToHex(c), wordToHex(d));
}

/*
 * A JavaScript implementation of the Secure Hash Algorithm, SHA-1, as defined in FIPS PUB 180-1 Version 2.1a
 * Copyright Paul Johnston 2000 - 2002. Other contributors: Greg Holt, Andrew Kepert, Ydnar, Lostinet
 * Distributed under the BSD License See http://pajhome.org.uk/crypt/md5 for details.
 */

/*
 * Configurable variables. You may need to tweak these to be compatible with the server-side, but the defaults
 * work in most cases.
 */
var hexcase = 0; /* hex output format. 0 - lowercase; 1 - uppercase */
var b64pad = ""; /* base-64 pad character. "=" for strict RFC compliance */
var chrsz = 8; /* bits per input character. 8 - ASCII; 16 - Unicode */

/*
 * These are the functions you'll usually want to call They take string arguments and return either hex or
 * base-64 encoded strings
 */

function hex_sha1(s) {
    return binb2hex(core_sha1(str2binb(s), s.length * chrsz));
}

function b64_sha1(s) {
    return binb2b64(core_sha1(str2binb(s), s.length * chrsz));
}

function str_sha1(s) {
    return binb2str(core_sha1(str2binb(s), s.length * chrsz));
}

function hex_hmac_sha1(key, data) {
    return binb2hex(core_hmac_sha1(key, data));
}

function b64_hmac_sha1(key, data) {
    return binb2b64(core_hmac_sha1(key, data));
}

function str_hmac_sha1(key, data) {
    return binb2str(core_hmac_sha1(key, data));
}

/*
 * Calculate the SHA-1 of an array of big-endian words, and a bit length
 */

function core_sha1(x, len) { /* append padding */
    x[len >> 5] |= 0x80 << (24 - len % 32);
    x[((len + 64 >> 9) << 4) + 15] = len;

    var w = Array(80);
    var a = 1732584193;
    var b = -271733879;
    var c = -1732584194;
    var d = 271733878;
    var e = -1009589776;

    for (var i = 0; i < x.length; i += 16) {
        var olda = a;
        var oldb = b;
        var oldc = c;
        var oldd = d;
        var olde = e;

        for (var j = 0; j < 80; j++) {
            if (j < 16) {
                w[j] = x[i + j];
            } else {
                w[j] = rol(w[j - 3] ^ w[j - 8] ^ w[j - 14] ^ w[j - 16], 1);
            }
            var t = safe_add(safe_add(rol(a, 5), sha1_ft(j, b, c, d)), safe_add(safe_add(e, w[j]), sha1_kt(j)));
            e = d;
            d = c;
            c = rol(b, 30);
            b = a;
            a = t;
        }

        a = safe_add(a, olda);
        b = safe_add(b, oldb);
        c = safe_add(c, oldc);
        d = safe_add(d, oldd);
        e = safe_add(e, olde);
    }
    return Array(a, b, c, d, e);

}

/*
 * Perform the appropriate triplet combination function for the current iteration
 */

function sha1_ft(t, b, c, d) {
    if (t < 20) {
        return (b & c) | ((~b) & d);
    }
    if (t < 40) {
        return b ^ c ^ d;
    }
    if (t < 60) {
        return (b & c) | (b & d) | (c & d);
    }
    return b ^ c ^ d;
}

/*
 * Determine the appropriate additive constant for the current iteration
 */

function sha1_kt(t) {
    return (t < 20) ? 1518500249 : (t < 40) ? 1859775393 : (t < 60) ? -1894007588 : -899497514;
}

/*
 * Calculate the HMAC-SHA1 of a key and some data
 */

function core_hmac_sha1(key, data) {
    var bkey = str2binb(key);
    if (bkey.length > 16) {
        bkey = core_sha1(bkey, key.length * chrsz);
    }

    var ipad = Array(16),
        opad = Array(16);
    for (var i = 0; i < 16; i++) {
        ipad[i] = bkey[i] ^ 0x36363636;
        opad[i] = bkey[i] ^ 0x5C5C5C5C;
    }

    var hash = core_sha1(ipad.concat(str2binb(data)), 512 + data.length * chrsz);
    return core_sha1(opad.concat(hash), 512 + 160);
}

/*
 * Add integers, wrapping at 2^32. This uses 16-bit operations internally to work around bugs in some JS
 * interpreters.
 */

function safe_add(x, y) {
    var lsw = (x & 0xFFFF) + (y & 0xFFFF);
    var msw = (x >> 16) + (y >> 16) + (lsw >> 16);
    return (msw << 16) | (lsw & 0xFFFF);
}

/*
 * Bitwise rotate a 32-bit number to the left.
 */

function rol(num, cnt) {
    return (num << cnt) | (num >>> (32 - cnt));
}

/*
 * Convert an 8-bit or 16-bit string to an array of big-endian words In 8-bit function, characters >255 have
 * their hi-byte silently ignored.
 */

function str2binb(str) {
    var bin = [];
    var mask = (1 << chrsz) - 1;
    for (var i = 0; i < str.length * chrsz; i += chrsz) {
        bin[i >> 5] |= (str.charCodeAt(i / chrsz) & mask) << (32 - chrsz - i % 32);
    }
    return bin;
}

/*
 * Convert an array of big-endian words to a string
 */

function binb2str(bin) {
    var str = "";
    var mask = (1 << chrsz) - 1;
    for (var i = 0; i < bin.length * 32; i += chrsz) {
        str += myStringFromCharCode((bin[i >> 5] >>> (32 - chrsz - i % 32)) & mask);
    }
    return str;
}

/*
 * Convert an array of big-endian words to a hex string.
 */

function binb2hex(binarray) {
    var hex_tab = hexcase ? "0123456789ABCDEF" : "0123456789abcdef";
    var str = "";
    for (var i = 0; i < binarray.length * 4; i++) {
        str += hex_tab.charAt((binarray[i >> 2] >> ((3 - i % 4) * 8 + 4)) & 0xF) + hex_tab.charAt((binarray[i >> 2] >> ((3 - i % 4) * 8)) & 0xF);
    }
    return str;
}

/*
 * Convert an array of big-endian words to a base-64 string
 */

function binb2b64(binarray) {
    var tab = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
    var str = "";
    for (var i = 0; i < binarray.length * 4; i += 3) {
        var triplet = (((binarray[i >> 2] >> 8 * (3 - i % 4)) & 0xFF) << 16) | (((binarray[i + 1 >> 2] >> 8 * (3 - (i + 1) % 4)) & 0xFF) << 8) | ((binarray[i + 2 >> 2] >> 8 * (3 - (i + 2) % 4)) & 0xFF);
        for (var j = 0; j < 4; j++) {
            if (i * 8 + j * 6 > binarray.length * 32) {
                str += b64pad;
            } else {
                str += tab.charAt((triplet >> 6 * (3 - j)) & 0x3F);
            }
        }
    }
    return str;
}

/**
 *
 * Secure Hash Algorithm (SHA256) http://www.webtoolkit.info/
 *
 * Original code by Angel Marin, Paul Johnston.
 *
 */

function SHA256(s) {

    var chrsz = 8;
    var hexcase = 0;

    function safe_add(x, y) {
        var lsw = (x & 0xFFFF) + (y & 0xFFFF);
        var msw = (x >> 16) + (y >> 16) + (lsw >> 16);
        return (msw << 16) | (lsw & 0xFFFF);
    }

    function S(X, n) {
        return (X >>> n) | (X << (32 - n));
    }

    function R(X, n) {
        return (X >>> n);
    }

    function Ch(x, y, z) {
        return ((x & y) ^ ((~x) & z));
    }

    function Maj(x, y, z) {
        return ((x & y) ^ (x & z) ^ (y & z));
    }

    function Sigma0256(x) {
        return (S(x, 2) ^ S(x, 13) ^ S(x, 22));
    }

    function Sigma1256(x) {
        return (S(x, 6) ^ S(x, 11) ^ S(x, 25));
    }

    function Gamma0256(x) {
        return (S(x, 7) ^ S(x, 18) ^ R(x, 3));
    }

    function Gamma1256(x) {
        return (S(x, 17) ^ S(x, 19) ^ R(x, 10));
    }

    function core_sha256(m, l) {
        var K = new Array(0x428A2F98, 0x71374491, 0xB5C0FBCF, 0xE9B5DBA5, 0x3956C25B, 0x59F111F1, 0x923F82A4, 0xAB1C5ED5, 0xD807AA98, 0x12835B01, 0x243185BE, 0x550C7DC3, 0x72BE5D74, 0x80DEB1FE, 0x9BDC06A7, 0xC19BF174, 0xE49B69C1, 0xEFBE4786, 0xFC19DC6, 0x240CA1CC, 0x2DE92C6F, 0x4A7484AA, 0x5CB0A9DC, 0x76F988DA, 0x983E5152, 0xA831C66D, 0xB00327C8, 0xBF597FC7, 0xC6E00BF3, 0xD5A79147, 0x6CA6351, 0x14292967, 0x27B70A85, 0x2E1B2138, 0x4D2C6DFC, 0x53380D13, 0x650A7354, 0x766A0ABB, 0x81C2C92E, 0x92722C85, 0xA2BFE8A1, 0xA81A664B, 0xC24B8B70, 0xC76C51A3, 0xD192E819, 0xD6990624, 0xF40E3585, 0x106AA070, 0x19A4C116, 0x1E376C08, 0x2748774C, 0x34B0BCB5, 0x391C0CB3, 0x4ED8AA4A, 0x5B9CCA4F, 0x682E6FF3, 0x748F82EE, 0x78A5636F, 0x84C87814, 0x8CC70208, 0x90BEFFFA, 0xA4506CEB, 0xBEF9A3F7, 0xC67178F2);
        var HASH = new Array(0x6A09E667, 0xBB67AE85, 0x3C6EF372, 0xA54FF53A, 0x510E527F, 0x9B05688C, 0x1F83D9AB, 0x5BE0CD19);
        var W = new Array(64);
        var a, b, c, d, e, f, g, h, i, j;
        var T1, T2;

        m[l >> 5] |= 0x80 << (24 - l % 32);
        m[((l + 64 >> 9) << 4) + 15] = l;

        for (var i = 0; i < m.length; i += 16) {
            a = HASH[0];
            b = HASH[1];
            c = HASH[2];
            d = HASH[3];
            e = HASH[4];
            f = HASH[5];
            g = HASH[6];
            h = HASH[7];

            for (var j = 0; j < 64; j++) {
                if (j < 16) {
                    W[j] = m[j + i];
                } else {
                    W[j] = safe_add(safe_add(safe_add(Gamma1256(W[j - 2]), W[j - 7]), Gamma0256(W[j - 15])), W[j - 16]);
                }

                T1 = safe_add(safe_add(safe_add(safe_add(h, Sigma1256(e)), Ch(e, f, g)), K[j]), W[j]);
                T2 = safe_add(Sigma0256(a), Maj(a, b, c));

                h = g;
                g = f;
                f = e;
                e = safe_add(d, T1);
                d = c;
                c = b;
                b = a;
                a = safe_add(T1, T2);
            }

            HASH[0] = safe_add(a, HASH[0]);
            HASH[1] = safe_add(b, HASH[1]);
            HASH[2] = safe_add(c, HASH[2]);
            HASH[3] = safe_add(d, HASH[3]);
            HASH[4] = safe_add(e, HASH[4]);
            HASH[5] = safe_add(f, HASH[5]);
            HASH[6] = safe_add(g, HASH[6]);
            HASH[7] = safe_add(h, HASH[7]);
        }
        return HASH;
    }

    function str2binb(str) {
        var bin = [];
        var mask = (1 << chrsz) - 1;
        for (var i = 0; i < str.length * chrsz; i += chrsz) {
            bin[i >> 5] |= (str.charCodeAt(i / chrsz) & mask) << (24 - i % 32);
        }
        return bin;
    }

    function Utf8Encode(string) {
        string = string.replace(/\r\n/g, "\n");
        var utftext = "";

        for (var n = 0; n < string.length; n++) {

            var c = string.charCodeAt(n);

            if (c < 128) {
                utftext += myStringFromCharCode(c);
            } else if ((c > 127) && (c < 2048)) {
                utftext += myStringFromCharCode((c >> 6) | 192);
                utftext += myStringFromCharCode((c & 63) | 128);
            } else {
                utftext += myStringFromCharCode((c >> 12) | 224);
                utftext += myStringFromCharCode(((c >> 6) & 63) | 128);
                utftext += myStringFromCharCode((c & 63) | 128);
            }

        }

        return utftext;
    }

    function binb2hex(binarray) {
        var hex_tab = hexcase ? "0123456789ABCDEF" : "0123456789abcdef";
        var str = "";
        for (var i = 0; i < binarray.length * 4; i++) {
            str += hex_tab.charAt((binarray[i >> 2] >> ((3 - i % 4) * 8 + 4)) & 0xF) + hex_tab.charAt((binarray[i >> 2] >> ((3 - i % 4) * 8)) & 0xF);
        }
        return str;
    }

    s = Utf8Encode(s);
    return binb2hex(core_sha256(str2binb(s), s.length * chrsz));

}


var Base64 = {

// public method for encoding
encode : function (input) {
    
    var bits = sjcl.codec.utf8String.toBits(input);
    return sjcl.codec.base64.fromBits(bits);
},

// public method for decoding
decode : function (input) {

    var bits =  sjcl.codec.base64.toBits(input);
    return sjcl.codec.utf8String.fromBits(bits);
}

}; //end base64

/**
 * @constructor
 */

function EncryptedCache() {

    // Internal Constants
    this.STORAGE_PREFIX = "__$WLEOC__";
    this.SALT_KEY = "__$WLEOC_SALT";
    this.CIPHER_KEY = "__$WLEOC_CIPHER";
    this.VERSION = "__$WLEOC_VERSION";

    this.CBK_KEY_DERIVATION_ITERATIONS = 1000;
    this.DPK_KEY_DERIVATION_ITERATIONS = 1000;
    this.DPK = null;
    this.keyCreationContext = null;

    if ((WL.Client.getEnvironment() === WL.Environment.ANDROID) || (WL.Client.getEnvironment() === WL.Environment.IPHONE) || (WL.Client.getEnvironment() === WL.Environment.IPAD)) {
        this[api.keygen] = new NativePBKDF2();
        this.encryptor = new NativeEncryptor();
    } else {
        this[api.keygen] = new WebBasedPBKDF2();
        this.encryptor = new WebBasedEncryptor();
    }

}

var OK = EncryptedCache.prototype["OK"] = 0;
var ERROR_NO_EOC = EncryptedCache.prototype["ERROR_NO_EOC"] = 1;
var ERROR_CREDENTIALS_MISMATCH = EncryptedCache.prototype["ERROR_CREDENTIALS_MISMATCH"] = 2;
var ERROR_EOC_TO_BE_DELETED = EncryptedCache.prototype["ERROR_EOC_TO_BE_DELETED"] = 3;
var ERROR_EOC_DELETED = EncryptedCache.prototype["ERROR_EOC_DELETED"] = 4;
var ERROR_UNSAFE_CREDENTIALS = EncryptedCache.prototype["ERROR_UNSAFE_CREDENTIALS"] = 5;
var ERROR_EOC_CLOSED = EncryptedCache.prototype["ERROR_EOC_CLOSED"] = 6;
var ERROR_NO_SUCH_KEY = EncryptedCache.prototype["ERROR_NO_SUCH_KEY"] = 7;
var ERROR_LOCAL_STORAGE_NOT_SUPPORTED = EncryptedCache.prototype["ERROR_LOCAL_STORAGE_NOT_SUPPORTED"] = 8;
var ERROR_KEY_CREATION_IN_PROGRESS = EncryptedCache.prototype["ERROR_KEY_CREATION_IN_PROGRESS"] = 9;
var ERROR_SECURE_RANDOM_GENERATOR_UNAVAILABLE = EncryptedCache.prototype["ERROR_SECURE_RANDOM_GENERATOR_UNAVAILABLE"] = 10;
var ERROR_COULD_NOT_GENERATE_KEY = EncryptedCache.prototype["ERROR_COULD_NOT_GENERATE_KEY"] = 11;
var ERROR_INVALID_PARAMETER = EncryptedCache.prototype["ERROR_INVALID_PARAMETER"] = 12;
var ERROR_UNKNOWN = EncryptedCache.prototype["ERROR_UNKNOWN"] = 13;
var ERROR_MIGRATION = EncryptedCache.prototype["ERROR_MIGRATION"] = 14;

EncryptedCache.prototype[api.close] = function(onCompleteHandler, onErrorHandler) {
    WL.Validators.validateArguments([WL.Validators.validateFunctionOrNull, WL.Validators.validateFunctionOrNull], arguments, 'WL.EncryptedCache.close');
    if (this.keyCreationContext) {
        if (onErrorHandler) {
            return onErrorHandler(ERROR_KEY_CREATION_IN_PROGRESS);
        } else {
            throw ERROR_KEY_CREATION_IN_PROGRESS;
        }
    }

    this.DPK = null;
    if (onCompleteHandler) {
        onCompleteHandler(OK);
    }
};

EncryptedCache.prototype[api.random] = function() {
    return myMathRandom();
};

EncryptedCache.prototype[api.secureRandom] = function(onCompleteHandler) {
    WL.Validators.validateArguments([WL.Validators.validateFunctionOrNull], arguments, 'WL.EncryptedCache.close');
    var serverUrl;
    
    if (WL["EnvProfile"]["isEnabled"](WL["EPField"]["WEB"]) && (WL.Client.getEnvironment() != WL.Env.MOBILE_WEB)) {
    	serverUrl = "/random";
    }
    else {
    	serverUrl = WL["StaticAppProps"]["APP_SERVICES_URL"] + "random";
    }
    
    new WLJSX.Ajax["WLRequest"](serverUrl, {
        onSuccess: function(response) {
            onCompleteHandler(response.responseText);
        },
        onFailure: function() {
            onCompleteHandler(ERROR_SECURE_RANDOM_GENERATOR_UNAVAILABLE);
        },
        timeout: WL["AppProp"]["WLCLIENT_TIMEOUT_IN_MILLIS"],
        method: "get",
        evalJSON: false
    });
};

/**
 *
 * @param credentials the secret identifier used to protect the cache
 * @param create_if_none whether to create the EOC if does not exist
 */
EncryptedCache.prototype[api.open] = function(credentials, create_if_none, onCompleteHandler, onErrorHandler) {
    try {
        WL.Validators.validateArguments(['string', 'boolean', WL.Validators.validateFunctionOrNull, WL.Validators.validateFunctionOrNull], arguments, 'WL.EncryptedCache.open');
    } catch (err) {
        console.log(err);
        if (onErrorHandler) {
            return onErrorHandler(ERROR_INVALID_PARAMETER);
        } else {
            throw ERROR_INVALID_PARAMETER;
        }
    }
    if ((credentials === undefined) || (credentials === '')) {
        console.log("Error: credentials is undefined or empty.");
        if (onErrorHandler) {
            return onErrorHandler(ERROR_INVALID_PARAMETER);
        } else {
            throw ERROR_INVALID_PARAMETER;
        }
    }

    if (this.DPK) {
        // close is really synchronous, but for consistency, we use callbacks
        var failureCode = null;
        this.close(function() {}, function(_failureCode) {
            failureCode = _failureCode;
        });
        if (failureCode !== null) {
            return onErrorHandler(failureCode);
        }
    }
    if (!localStorage) {
        if (onErrorHandler) {
            return onErrorHandler(ERROR_LOCAL_STORAGE_NOT_SUPPORTED);
        } else {
            throw ERROR_LOCAL_STORAGE_NOT_SUPPORTED;
        }
    }
    if (this.keyCreationContext) {
        if (onErrorHandler) {
            return onErrorHandler(ERROR_KEY_CREATION_IN_PROGRESS);
        } else {
            throw ERROR_KEY_CREATION_IN_PROGRESS;
        }
    }
    var salt = localStorage.getItem(this.SALT_KEY);
    if (salt === null) {
        if (!create_if_none) {
            if (onErrorHandler) {
                return onErrorHandler(ERROR_NO_EOC);
            } else {
                throw ERROR_NO_EOC;
            }
        }

        // TODO: strengthen salt
        this.keyCreationContext = {
            credentials: credentials,
            salt: "" + this[api.random](),
            onCompleteHandler: onCompleteHandler,
            onErrorHandler: onErrorHandler
        };

        var me = this;

        // Create and store pseudorandom cipher-key
        // TODO: Strengthen key creation
        this[api.secureRandom](function(response) {
            if (response == ERROR_SECURE_RANDOM_GENERATOR_UNAVAILABLE) {
                var onErrorHandler = me.keyCreationContext.onErrorHandler;
                me.keyCreationContext = null;
                this.keyCreationContext = null;
                if (onErrorHandler) {
                    return onErrorHandler(ERROR_SECURE_RANDOM_GENERATOR_UNAVAILABLE);
                } else {
                    throw ERROR_SECURE_RANDOM_GENERATOR_UNAVAILABLE;
                }
            }
            var random = response;
            me[api.keygen].deriveKey(random, me.keyCreationContext.salt, me.DPK_KEY_DERIVATION_ITERATIONS, 32, function(key) {
                me.keyCreationContext.DPK = key;
                me[api.keygen].deriveKey(me.keyCreationContext.credentials, me.keyCreationContext.salt, me.CBK_KEY_DERIVATION_ITERATIONS, 32, function(cbkKey) {
                    // encrypt DPK with CBK
                    var iv = [MD5("" + me[api.random]())];
                    me.encryptor.rawEncrypt(me.keyCreationContext.DPK, cbkKey, iv[0], function(encryptedDPK) {
                        //Result Callback
                        // Store
                        // encryption
                        // data
                        localStorage.setItem(me.CIPHER_KEY, encryptedDPK);
                        localStorage.setItem(me.SALT_KEY, me.keyCreationContext.salt);
                        
                        // Creating a new cache using the correct encryption so we set the version
                        localStorage.setItem(me.VERSION, "1");

                        var onCompleteHandler = me.keyCreationContext.onCompleteHandler;

                        // We
                        // are
                        // ready
                        me.DPK = me.keyCreationContext.DPK;
                        me.keyCreationContext = null;
                        this.keyCreationContext = null;

                        if (onCompleteHandler) {
                            onCompleteHandler(OK);
                        }
                    }, function(err) {
                        console.log(err);
                        var onErrorHandler = me.keyCreationContext.onErrorHandler;
                        me.keyCreationContext = null;
                        this.keyCreationContext = null;
                        if (onErrorHandler) {
                            onErrorHandler(ERROR_COULD_NOT_GENERATE_KEY);
                        } else {
                            throw ERROR_COULD_NOT_GENERATE_KEY;
                        }
                    });
                }, function(err) {
                    console.log(err);
                    var onErrorHandler = me.keyCreationContext.onErrorHandler;
                    me.keyCreationContext = null;
                    this.keyCreationContext = null;
                    if (onErrorHandler) {
                        onErrorHandler(ERROR_COULD_NOT_GENERATE_KEY);
                    } else {
                        throw ERROR_COULD_NOT_GENERATE_KEY;
                    }
                });
            }, function(err) {
                console.log(err);
                var onErrorHandler = me.keyCreationContext.onErrorHandler;
                me.keyCreationContext = null;
                this.keyCreationContext = null;
                if (onErrorHandler) {
                    onErrorHandler(ERROR_COULD_NOT_GENERATE_KEY);
                } else {
                    throw ERROR_COULD_NOT_GENERATE_KEY;
                }
            });
        });

    } else {
        // EOC already initialized, decrypt DPK
        this.keyCreationContext = {
            onCompleteHandler: onCompleteHandler,
            onErrorHandler: onErrorHandler
        };

        var me = this;

        this[api.keygen].deriveKey(credentials, salt, me.CBK_KEY_DERIVATION_ITERATIONS, 32, function(cbkKey) {
            var onCompleteHandler = me.keyCreationContext.onCompleteHandler;
            var onErrorHandler = me.keyCreationContext.onErrorHandler;
            me.keyCreationContext = null;
            this.keyCreationContext = null;

            // decrypt DPK with CBK
            var encryptedDPK = localStorage.getItem(me.CIPHER_KEY);

            me.encryptor.rawDecrypt(encryptedDPK, cbkKey, function(decryptedDPK) {
                me.DPK = decryptedDPK;
                
                // Check to see if EOC needs to be migrated
                var version = localStorage.getItem(me.VERSION);
                var platform = WL.Client.getEnvironment();
                if( (platform === 'iphone' || platform === 'ipad') && (typeof version === 'undefined' || typeof version !== 'string') ){
                	reEncryptStorage(onCompleteHandler, onErrorHandler, me, cbkKey);
                }else{
	                if (onCompleteHandler) {
	                    onCompleteHandler(OK);
	                }
                }
            }, function(err) {
                console.log(err);
                if (onErrorHandler) {
                    onErrorHandler(ERROR_CREDENTIALS_MISMATCH);
                } else {
                    throw ERROR_CREDENTIALS_MISMATCH;
                }
            });

        }, function(err) {
            console.log(err);
            var onErrorHandler = me.keyCreationContext.onErrorHandler;
            me.keyCreationContext = null;
            this.keyCreationContext = null;
            if (onErrorHandler) {
                onErrorHandler(ERROR_COULD_NOT_GENERATE_KEY);
            } else {
                throw ERROR_COULD_NOT_GENERATE_KEY;
            }
        });

    }
};

function reEncryptStorage(onCompleteHandler, onErrorHandler, encObj, cbkKey){
	
	// Use the same function for handling any errors we receive
	function reEncryptError(err) {
        console.log(err);
        if (onErrorHandler) {
            onErrorHandler(ERROR_MIGRATION);
        } else {
            throw ERROR_MIGRATION;
        }
	}
	
	var iv = [MD5("" + encObj[api.random]())];
    
	encObj.encryptor.rawEncrypt(encObj.DPK, cbkKey, iv[0], function(encryptedDPK) {
		localStorage.setItem(encObj.CIPHER_KEY, encryptedDPK);
		reEncryptKeys(onCompleteHandler, reEncryptError, encObj);
	}, reEncryptError);
			
}

function reEncryptKeys(onCompleteHandler, onErrorHandler, encObj){
	
	// Begin here
	var allKeys = [];
	var currentKey = "";
	
	// Collect all of the keys
	for (var i = localStorage.length - 1; i >= 0; i--) {
        var key = localStorage.key(i);
        
        if (key.indexOf(encObj.STORAGE_PREFIX) === 0) {
        	allKeys.push(key);
        }
    }
	
	encryptNextKey();
	
	// Encrypt the next key
	function encryptNextKey(){
		
		if(allKeys.length > 0){
			currentKey = allKeys.pop();
			readUsingHashedKey(currentKey, encObj, reEncryptKeyData, onErrorHandler);
		}else{
			// Save the version to the local storage
			localStorage.setItem(encObj.VERSION, "1");
			
			if (onCompleteHandler) {
                onCompleteHandler(OK);
            }
		}
	}
	
	// After decrypting data, reEncrypt it
	function reEncryptKeyData( value ){
		writeUsingHashedKey(currentKey, value, encObj, encryptNextKey, onErrorHandler);	
    }
}

function readUsingHashedKey(key, encObj, onCompleteHandler, onErrorHandler){
	var encryptedData = localStorage.getItem(key);
    encObj.encryptor.rawDecrypt(encryptedData, encObj.DPK, onCompleteHandler, onErrorHandler);
}

function writeUsingHashedKey(key, value, encObj, onCompleteHandler, onErrorHandler){
	var storagePrefix = encObj.STORAGE_PREFIX;
	var noPrefixKey = key.replace(storagePrefix,'');
    var iv = [MD5(noPrefixKey + encObj[api.random]())];
    
    encObj.encryptor.rawEncrypt(value, encObj.DPK, iv[0], function(encryptedData) {
        localStorage.setItem(storagePrefix + noPrefixKey, encryptedData);
        onCompleteHandler();
    }, onErrorHandler);
}

EncryptedCache.prototype[api.destroy] = function(onCompleteHandler, onErrorHandler) {
    WL.Validators.validateArguments([WL.Validators.validateFunctionOrNull, WL.Validators.validateFunctionOrNull], arguments, 'WL.EncryptedCache.destroy');
    // close is really synchronous, but for consistency, we use callbacks
    var failureCode = null;
    this.close(function() {}, function(_failureCode) {
        failureCode = _failureCode;
    });
    if (failureCode !== null) {
        if (onErrorHandler) {
            return onErrorHandler(failureCode);
        } else {
            throw failureCode;
        }
    }

    localStorage.removeItem(this.SALT_KEY);
    localStorage.removeItem(this.CIPHER_KEY);

    for (var i = localStorage.length - 1; i >= 0; i--) {
        var key = localStorage.key(i);
        // Remove all keys
        if (key.indexOf(this.STORAGE_PREFIX) === 0) {
            localStorage.removeItem(key);
        }
    }
    if (onCompleteHandler) {
    	onCompleteHandler(OK);
    }
};

EncryptedCache.prototype[api.changeCredentials] = function(new_credentials, onCompleteHandler, onErrorHandler) {
    try {
        WL.Validators.validateArguments(['string', WL.Validators.validateFunctionOrNull, WL.Validators.validateFunctionOrNull], arguments, 'WL.EncryptedCache.changeCredentials');
    } catch (err) {
        console.log(err);
        if (onErrorHandler) {
            return onErrorHandler(ERROR_INVALID_PARAMETER);
        } else {
            throw ERROR_INVALID_PARAMETER;
        }
    }

    if ((new_credentials === undefined) || (new_credentials === '')) {
        console.log("Error: new_credentials is undefined or empty.");
        if (onErrorHandler) {
            return onErrorHandler(ERROR_INVALID_PARAMETER);
        } else {
            throw ERROR_INVALID_PARAMETER;
        }
    }

    if (this.keyCreationContext) {
        if (onErrorHandler) {
            return onErrorHandler(ERROR_KEY_CREATION_IN_PROGRESS);
        } else {
            throw ERROR_KEY_CREATION_IN_PROGRESS;
        }
    }
    if (!this.DPK) {
        if (onErrorHandler) {
            return onErrorHandler(ERROR_EOC_CLOSED);
        } else {
            throw ERROR_EOC_CLOSED;
        }
    }

    // TODO: strengthen salt
    this.keyCreationContext = {
        credentials: new_credentials,
        salt: "" + this[api.random](),
        onCompleteHandler: onCompleteHandler,
        onErrorHandler: onErrorHandler
    };

    var me = this;

    this[api.keygen].deriveKey(this.keyCreationContext.credentials, this.keyCreationContext.salt, this.CBK_KEY_DERIVATION_ITERATIONS, 32, function(cbkKey) {
        // encrypt DPK with CBK
        var iv = [MD5("" + me[api.random]())];
        me.encryptor.rawEncrypt(me.DPK, cbkKey, iv[0], function(encryptedDPK) {
            // Store encryption data
            localStorage.setItem(me.CIPHER_KEY, encryptedDPK);
            localStorage.setItem(me.SALT_KEY, me.keyCreationContext.salt);

            var onCompleteHandler = me.keyCreationContext.onCompleteHandler;

            // We are ready
            me.keyCreationContext = null;
            this.keyCreationContext = null;

            if (onCompleteHandler) {
                onCompleteHandler(OK);
            }
        }, function(err) {
            console.log(err);
            var onErrorHandler = me.keyCreationContext.onErrorHandler;
            me.keyCreationContext = null;
            this.keyCreationContext = null;
            if (onErrorHandler) {
                onErrorHandler(ERROR_COULD_NOT_GENERATE_KEY);
            } else {
                throw ERROR_COULD_NOT_GENERATE_KEY;
            }
        });

        },
    function(err) {
        console.log(err);
        var onErrorHandler = me.keyCreationContext.onErrorHandler;
        me.keyCreationContext = null;
        this.keyCreationContext = null;
        if (onErrorHandler) {
            onErrorHandler(ERROR_COULD_NOT_GENERATE_KEY);
        } else {
            throw ERROR_COULD_NOT_GENERATE_KEY;
        }
    });
};

EncryptedCache.prototype[api.write] = function(key, data, onCompleteHandler, onErrorHandler) {
    try {
        WL.Validators.validateArguments(['string', WL.Validators.validateStringOrNull, WL.Validators.validateFunctionOrNull, WL.Validators.validateFunctionOrNull], arguments, 'WL.EncryptedCache.write');
    } catch (err) {
        console.log(err);
        if (onErrorHandler) {
            return onErrorHandler(ERROR_INVALID_PARAMETER);
        } else {
            throw ERROR_INVALID_PARAMETER;
        }
    }

    if ((key === undefined) || (key === '')) {
        console.log("Error: key is undefined or empty.");
        if (onErrorHandler) {
            return onErrorHandler(ERROR_INVALID_PARAMETER);
        } else {
            throw ERROR_INVALID_PARAMETER;
        }
    }

    if (!this.DPK) {
        return onErrorHandler(ERROR_EOC_CLOSED);
    }

    // Key names are hashed using SHA256
    key = SHA256(key);

    if (data === null) {
        localStorage.removeItem(this.STORAGE_PREFIX + key);
        if (onCompleteHandler) {
        	return onCompleteHandler(OK);
        }
        return;
    }

    // Use the key name as an additional factor when creating the IV
    var iv = [MD5(key + this[api.random]())];
    var storagePrefix = this.STORAGE_PREFIX;
    var base64encoded_data = Base64.encode(data);
    
    this.encryptor.rawEncrypt(base64encoded_data, this.DPK, iv[0], function(encryptedData) {
        //Result Callback
        localStorage.setItem(storagePrefix + key, encryptedData);
        if (onCompleteHandler) {
        	onCompleteHandler(OK);
        }
    }, onErrorHandler);

};

EncryptedCache.prototype[api.remove] = function(key, onCompleteHandler, onErrorHandler) {
    try {
        WL.Validators.validateArguments(['string', WL.Validators.validateFunctionOrNull, WL.Validators.validateFunctionOrNull], arguments, 'WL.EncryptedCache.remove');
    } catch (err) {
        console.log(err);
        if (onErrorHandler) {
            return onErrorHandler(ERROR_INVALID_PARAMETER);
        } else {
            throw ERROR_INVALID_PARAMETER;
        }
    }

    if ((key === undefined) || (key === '')) {
        console.log("Error: key is undefined or empty.");
        if (onErrorHandler) {
            return onErrorHandler(ERROR_INVALID_PARAMETER);
        } else {
            throw ERROR_INVALID_PARAMETER;
        }
    }

    if (!this.DPK) {
        if (onErrorHandler) {
            return onErrorHandler(ERROR_EOC_CLOSED);
        } else {
            throw ERROR_EOC_CLOSED;
        }
    }
    this.write(key, null, onCompleteHandler, onErrorHandler);
};

EncryptedCache.prototype[api.read] = function(key, onCompleteHandler, onErrorHandler) {
    try {
        WL.Validators.validateArguments(['string', WL.Validators.validateFunctionOrNull, WL.Validators.validateFunctionOrNull], arguments, 'WL.EncryptedCache.read');
    } catch (err) {
        console.log(err);
        if (onErrorHandler) {
            return onErrorHandler(ERROR_INVALID_PARAMETER);
        } else {
            throw ERROR_INVALID_PARAMETER;
        }
    }

    if ((key === undefined) || (key === '')) {
        console.log("Error: key is undefined or empty.");
        if (onErrorHandler) {
            return onErrorHandler(ERROR_INVALID_PARAMETER);
        } else {
            throw ERROR_INVALID_PARAMETER;
        }
    }


    if (!this.DPK) {
        if (onErrorHandler) {
            return onErrorHandler(ERROR_EOC_CLOSED);
        } else {
            throw ERROR_EOC_CLOSED;
        }
    }

    // Key names are hashed using SHA256
    key = SHA256(key);

    var encryptedData = localStorage.getItem(this.STORAGE_PREFIX + key);
    if (encryptedData === null) {
    	if (onCompleteHandler) {
    		return onCompleteHandler(null);
    	}
    	return;
    }

    function base64_decode_callback( base64encoded_data ){
        var plainText = Base64.decode(base64encoded_data);
        if (onCompleteHandler) {
        	onCompleteHandler(plainText);
        }
    }

    this.encryptor.rawDecrypt(encryptedData, this.DPK, base64_decode_callback, onErrorHandler);
};

/** @constructor */

function NativeEncryptor() {}

/**
 *
 * @param encryptedData expect a string like "plainText"
 * @param key
 * @returns {String} of the form "{"ct":"encrypted_data","iv":"some_iv"}"
 */
NativeEncryptor.prototype.rawEncrypt = function(plainText, key, iv, onCompleteHandler, onErrorHandler) {

    try {
        WL.Validators.validateArguments(['string', 'string', 'object', 'function', 'function'], arguments, 'NativeEncryptor.prototype.rawEncrypt');
    } catch (err) {
        console.log(err);
        if (onErrorHandler) {
            return onErrorHandler(ERROR_INVALID_PARAMETER);
        } else {
            throw ERROR_INVALID_PARAMETER;
        }
    }

    var params = {
        value: plainText,
        key: key,
        iv: a2h(iv)
    };

    if ((!params.key) || (!params.value) || (!params.iv) || (params.key === null) || (params.value === null) || (params.iv === null) || (params.key.length < 1) || (params.value.length < 1 || (params.iv.length < 1))) {
        if (onErrorHandler) {
            return onErrorHandler(ERROR_INVALID_PARAMETER);
        } else {
            throw ERROR_INVALID_PARAMETER;
        }
    }

    // function wrapper prevent returning sensitive data from the cordova plugin
    cordova.exec(onCompleteHandler, function() {
        onErrorHandler();
    }, "SecurityPlugin", "encrypt", [params.key, params.value, params.iv]);
};

/**
 *
 * @param encryptedData expect a string like "{"ct":"encrypted_data","iv":"some_iv"}"
 * @param key
 */
NativeEncryptor.prototype.rawDecrypt = function(encryptedData, key, onCompleteHandler, onErrorHandler) {

    try {
        WL.Validators.validateArguments(['string', 'string', 'function', 'function'], arguments, 'NativeEncryptor.prototype.rawDecrypt');
    } catch (err) {
        console.log(err);
        if (onErrorHandler) {
            return onErrorHandler(ERROR_INVALID_PARAMETER);
        } else {
            throw ERROR_INVALID_PARAMETER;
        }
    }

    var encryptedDataJSON = JSON.parse(encryptedData);
    var iv = encryptedDataJSON.iv;

    var params = {
        value: encryptedDataJSON.ct,
        key: key,
        iv: iv
    };

    if ((!params.key) || (!params.value) || (!params.iv) || (params.key === null) || (params.value === null) || (params.iv === null) || (params.key.length < 1) || (params.value.length < 1) || (params.iv.length < 1)) {
        if (onErrorHandler) {
            return onErrorHandler(ERROR_INVALID_PARAMETER);
        } else {
            throw ERROR_INVALID_PARAMETER;
        }
    }

    // function wrapper prevent returning sensitive data from the cordova plugin
    cordova.exec(onCompleteHandler, function(err) {
        console.log(err);
        onErrorHandler(ERROR_UNKNOWN);
    }, "SecurityPlugin", "decrypt", [params.key, encryptedDataJSON.ct, params.iv]);
};

/** @constructor */

function NativePBKDF2() {}

// Starts the computation
NativePBKDF2.prototype.deriveKey = function(password, salt, num_iterations, num_bytes, onCompleteHandler, onErrorHandler) {

    try {
        WL.Validators.validateArgument('string', password, 'NativePBKDF2.prototype.deriveKey');
        WL.Validators.validateArgument('string', salt, 'NativePBKDF2.prototype.deriveKey');
        WL.Validators.validateArgument(function(arg) {
            if (typeof arg !== 'function') {
                this.logAndThrow("Invalid value '" + WLJSX.Object.toJSON(arg) + "' (" + (typeof arg) + "), expected type 'function' for parameter onCompleteHandler.", 'NativePBKDF2.prototype.deriveKey');
            }
        }, onCompleteHandler, "");
        WL.Validators.validateArgument(function(arg) {
            if (typeof arg !== 'function') {
                this.logAndThrow("Invalid value '" + WLJSX.Object.toJSON(arg) + "' (" + (typeof arg) + "), expected type 'function' for parameter onErrorHandler.", 'NativePBKDF2.prototype.deriveKey');
            }
        }, onErrorHandler, "");
        // do the integer validation ourselves since WL.Validators does not provide this type of validator (yet)
        if (!((typeof num_iterations === 'number') && (Math.ceil(num_iterations) === Math.floor(num_iterations)))) {
            throw "Error: Invalid invocation of method NativePBKDF2.prototype.deriveKey; Invalid parameter type for argument 'num_iterations'";
        }
        if (!((typeof num_bytes === 'number') && (Math.ceil(num_bytes) === Math.floor(num_bytes)))) {
            throw "Error: Invalid invocation of method NativePBKDF2.prototype.deriveKey; Invalid parameter type for argument 'num_bytes'";
        }
    } catch (err) {
        console.log(err);
        if (onErrorHandler) {
            return onErrorHandler(ERROR_INVALID_PARAMETER);
        } else {
            throw ERROR_INVALID_PARAMETER;
        }
    }

    var params = {
        pass: password,
        salt: salt,
        iter: num_iterations,
        len: num_bytes
    };

    if ((!params.pass) || (params.pass === null) || (!params.salt) || (params.salt === null) || (!params.iter) || (params.iter === null) || (!params.len) || (params.len === null) || (params.salt.length < 1) || (parseInt(params.iter,10) < 0)) {
        return onErrorHandler(ERROR_INVALID_PARAMETER);
    }

    cordova.exec(onCompleteHandler, function(err) {
        console.log(err);
        onErrorHandler(ERROR_UNKNOWN);
    }, "SecurityPlugin", "keygen", [params.pass, params.salt, params.iter, params.len]);
};

function WebBasedEncryptor() {}

WebBasedEncryptor.prototype.rawEncrypt = function(plainText, key, iv, onCompleteHandler, onErrorHandler) {

    var result = null;
    try {
        // TODO: validate key is an 8-element array
        iv = sjcl.codec.hex.toBits(a2h(iv));
        result = sjcl.json.encrypt(key, plainText, {
            iv: iv
        });
    } catch (err) {
        console.log(err);
        if (onErrorHandler) {
            return onErrorHandler(ERROR_UNKNOWN);
        } else {
            throw ERROR_UNKNOWN;
        }
    }

    if (onCompleteHandler) {
        onCompleteHandler(result);
    }
};

WebBasedEncryptor.prototype.rawDecrypt = function(encryptedData, key, onCompleteHandler, onErrorHandler) {

    var result = null;
    try {
        result = sjcl.json.decrypt(key, encryptedData);
    } catch (err) {
        console.log(err);
        if (onErrorHandler) {
            return onErrorHandler(ERROR_UNKNOWN);
        } else {
            throw ERROR_UNKNOWN;
        }
    }

    if (onCompleteHandler) {
        onCompleteHandler(result);
    }
};

/** @constructor */

function WebBasedPBKDF2() {}

WebBasedPBKDF2.prototype.deriveKey = function(password, salt, num_iterations, num_bytes, onCompleteHandler, onErrorHandler) {
    var key = null;
    try {
        key = sjcl.misc.pbkdf2(password, salt, num_iterations, num_bytes * 8); // TODO * 8
    } catch (err) {
        console.log(err);
        if (onErrorHandler) {
            return onErrorHandler(ERROR_UNKNOWN);
        } else {
            throw ERROR_UNKNOWN;
        }
    }

    if (onCompleteHandler) {
        onCompleteHandler(binb2hex(key));
    }
};

window["WL"]["EncryptedCache"] = new EncryptedCache();
